@@ -34,12 +34,11 @@ from itertools import chain
 from pathlib import Path
 from textwrap import dedent
 from types import FunctionType, ModuleType
-from typing import Dict, Iterator, List, Optional, Set, Tuple
+from typing import Dict, Iterator, List, Optional, Set, Tuple, Union
 
 import pydantic
 import pytest
 import yaml
-from pytest import CaptureFixture
 from tornado import ioloop
 
 import inmanta.ast
@@ -59,70 +58,46 @@ if typing.TYPE_CHECKING:
     from inmanta.agent.io.local import IOBase
 
 from .handler import DATA
+from .parameters import (
+    inm_install_mode,
+    inm_mod_in_place,
+    inm_mod_repo,
+    inm_no_load_plugins,
+    inm_venv,
+)
+from .test_parameter import ParameterNotSetException, TestParameterRegistry
+
+try:
+    """
+    Those classes are only used in type annotation, but the import doesn't work
+    in python 3.6.  So we simply catch the error and ignore it.
+    """
+    from pytest import CaptureFixture, OptionGroup, Parser
+except ImportError:
+    pass
 
 CURDIR = os.getcwd()
 LOGGER = logging.getLogger()
 SYS_EXECUTABLE = sys.executable
 
 
-option_to_env = {
-    "inm_venv": "INMANTA_TEST_ENV",
-    "inm_module_repo": "INMANTA_MODULE_REPO",
-    "inm_install_mode": "INMANTA_INSTALL_MODE",
-    "inm_no_load_plugins": "INMANTA_TEST_NO_LOAD_PLUGINS",
-}
+def pytest_addoption(parser: "Parser") -> None:
+    for group_name, parameters in TestParameterRegistry.test_parameter_groups().items():
+        group: Union["Parser", "OptionGroup"]
+        if group_name is None:
+            group = parser
+        else:
+            group = parser.getgroup(group_name)
 
+        for param in parameters:
+            kwargs: Dict[str, object] = dict(
+                action=param.action,
+                help=param.help,
+            )
+            if param.choices is not None:
+                kwargs["choices"] = param.choices
 
-def pytest_addoption(parser) -> None:
-    group = parser.getgroup("inmanta", "inmanta module testing plugin")
-    group.addoption(
-        "--venv",
-        dest="inm_venv",
-        help="folder in which to place the virtual env for tests (will be shared by all tests), overrides INMANTA_TEST_ENV. "
-        "This options depends on symlink support. This does not work on all windows versions. "
-        "On windows 10 you need to run pytest in an admin shell. "
-        "Using a fixed virtual environment can speed up running the tests.",
-    )
-    group.addoption(
-        "--use-module-in-place",
-        action="store_true",
-        help="tell pytest-inmanta to run with the module in place, useful for debugging. "
-        "Makes inmanta add the parent directory of your module directory to it's directory path, instead of copying your "
-        "module to a temporary libs directory. "
-        "It allows testing the current module against specific versions of dependent modules. "
-        "Using this option can speed up the tests, because the module dependencies are not downloaded multiple times.",
-    )
-    group.addoption(
-        "--module_repo",
-        dest="inm_module_repo",
-        action="append",
-        help="location to download modules from, overrides INMANTA_MODULE_REPO."
-        "Can be specified multiple times to add multiple locations",
-    )
-    group.addoption(
-        "--install_mode",
-        dest="inm_install_mode",
-        help="Install mode for modules downloaded during this test, overrides INMANTA_INSTALL_MODE.",
-        choices=module.INSTALL_OPTS,
-    )
-    group.addoption(
-        "--no_load_plugins",
-        action="store_true",
-        dest="inm_no_load_plugins",
-        help="Don't load plugins in the Project class. Overrides INMANTA_TEST_NO_LOAD_PLUGINS."
-        "The value of INMANTA_TEST_NO_LOAD_PLUGINS environment variable has to be a non-empty string to not load plugins."
-        "When not using this option during the testing of plugins with the `project.get_plugin_function` method, "
-        "it's possible that the module's `plugin/__init__.py` is loaded multiple times, "
-        "which can cause issues when it has side effects, as they are executed multiple times as well.",
-    )
-
-
-def get_opt_or_env_or(config, key: str, default: str) -> str:
-    if config.getoption(key):
-        return config.getoption(key)
-    if option_to_env[key] in os.environ:
-        return os.environ[option_to_env[key]]
-    return default
+            group.addoption(param.argument, **kwargs)
 
 
 def get_module() -> typing.Tuple[module.Module, str]:
@@ -165,7 +140,7 @@ def inmanta_plugins(
 
 @pytest.fixture()
 def project(
-    project_shared: "Project", capsys: CaptureFixture
+    project_shared: "Project", capsys: "CaptureFixture"
 ) -> typing.Iterator["Project"]:
     DATA.clear()
     project_shared.clean()
@@ -176,12 +151,12 @@ def project(
 
 @pytest.fixture()
 def project_no_plugins(
-    project_shared_no_plugins, capsys: CaptureFixture
+    project_shared_no_plugins: "Project", capsys: "CaptureFixture"
 ) -> typing.Iterator["Project"]:
     warnings.warn(
         DeprecationWarning(
             "The project_no_plugins fixture is deprecated in favor of the %s environment variable."
-            % option_to_env["inm_no_load_plugins"]
+            % inm_no_load_plugins.environment_variable
         )
     )
     DATA.clear()
@@ -259,9 +234,7 @@ def project_factory(request: pytest.FixtureRequest) -> typing.Callable[[], "Proj
     test_project_dir = tempfile.mkdtemp()
     os.mkdir(os.path.join(test_project_dir, "libs"))
 
-    repo_options = get_opt_or_env_or(
-        request.config, "inm_module_repo", "https://github.com/inmanta/"
-    )
+    repo_options = inm_mod_repo.resolve(request.config)
     repos: typing.Sequence[object] = get_project_repos(
         chain.from_iterable(
             repo.split(" ")
@@ -271,14 +244,17 @@ def project_factory(request: pytest.FixtureRequest) -> typing.Callable[[], "Proj
         )
     )
 
-    install_mode = get_opt_or_env_or(request.config, "inm_install_mode", "release")
+    install_mode = inm_install_mode.resolve(request.config)
 
     modulepath = ["libs"]
-    in_place = request.config.getoption("--use-module-in-place")
+    in_place = inm_mod_in_place.resolve(request.config)
     if in_place:
         modulepath.append(str(Path(CURDIR).parent))
 
-    env_override = get_opt_or_env_or(request.config, "inm_venv", None)
+    try:
+        env_override: Optional[str] = str(inm_venv.resolve(request.config))
+    except ParameterNotSetException:
+        env_override = None
     env_dir = os.path.join(test_project_dir, ".env")
     if env_override and not os.path.isdir(env_override):
         raise Exception(f"Specified venv {env_override} does not exist")
@@ -300,17 +276,16 @@ def project_factory(request: pytest.FixtureRequest) -> typing.Callable[[], "Proj
             "repo": repos,
             "modulepath": modulepath,
             "downloadpath": "libs",
-            "install_mode": install_mode,
+            "install_mode": install_mode.value,
         }
         yaml.dump(metadata, fd)
 
     ensure_current_module_install(os.path.join(test_project_dir, "libs"), in_place)
 
     def create_project(**kwargs: object):
+        load_plugins = not inm_no_load_plugins.resolve(request.config)
         extended_kwargs: typing.Dict[str, object] = {
-            "load_plugins": not get_opt_or_env_or(
-                request.config, "inm_no_load_plugins", False
-            ),
+            "load_plugins": load_plugins,
             "env_path": env_dir,
             **kwargs,
         }
@@ -655,7 +630,7 @@ class Project:
         self._should_load_plugins: typing.Optional[bool] = load_plugins
         self._plugins: typing.Optional[typing.Dict[str, FunctionType]] = None
         self._load()
-        self._capsys: typing.Optional[CaptureFixture] = None
+        self._capsys: typing.Optional["CaptureFixture"] = None
         self.ctx: typing.Optional[HandlerContext] = None
         self._handlers: typing.Set[ResourceHandler] = set()
         config.Config.load_config()
@@ -672,7 +647,7 @@ class Project:
 
         sys.executable = compiler_executable
 
-    def init(self, capsys: CaptureFixture) -> None:
+    def init(self, capsys: "CaptureFixture") -> None:
         self._stdout = None
         self._stderr = None
         self._capsys = capsys
